#!/usr/bin/env python

import os
import sys
import time
import logging
from datetime import timedelta, datetime, date

import elasticsearch
#from curator import *
import curator

try:
    from logging import NullHandler
except ImportError:
    from logging import Handler

    class NullHandler(Handler):
        def emit(self, record):
            pass

__version__ = '2.0.0'

# Elasticsearch versions supported
version_max  = (2, 0, 0)
version_min = (1, 0, 0)
        
logger = logging.getLogger(__name__)

DEFAULT_ARGS = {
    'host': 'localhost',
    'url_prefix': '',
    'port': 9200,
    'auth': None,
    'ssl': False,
    'timeout': 30,
    'prefix': 'logstash-',
    'suffix': '',
    'curation_style': 'time',
    'time_unit': 'days',
    'max_num_segments': 2,
    'dry_run': False,
    'debug': False,
    'log_level': 'INFO',
    'logformat': 'Default',
    'all_indices': False,
    'show_indices': False,
    'snapshot_prefix': 'curator-',
    'wait_for_completion': True,
    'ignore_unavailable': False,
    'include_global_state': True,
    'partial': False,
}

DATEMAP = {
    'months': '%Y.%m',
    'weeks': '%Y.%W',
    'days': '%Y.%m.%d',
    'hours': '%Y.%m.%d.%H',
}

def add_common_args(subparser):
    """Add common arguments here to reduce redundancy and line count"""
    subparser.add_argument('--timestring', help="Python strftime string to match your index definition, e.g. 2014.07.15 would be %%Y.%%m.%%d", type=str, default=None)
    subparser.add_argument('--prefix', help='Define a prefix. Index name = PREFIX + TIMESTRING + SUFFIX. Default: logstash-', default=DEFAULT_ARGS['prefix'])
    subparser.add_argument('--suffix', help='Define a suffix. Index name = PREFIX + TIMESTRING + SUFFIX. Default: Empty', default=DEFAULT_ARGS['suffix'])
    subparser.add_argument('--time-unit', dest='time_unit', action='store', help='Unit of time to reckon by: [hours|days|weeks|months] Default: days',
                        default=DEFAULT_ARGS['time_unit'], type=str)
    subparser.add_argument('--exclude-pattern', help='Exclude indices matching provided pattern, e.g. 2014.06.08', type=str, default=None)
    
def make_parser():
    """ Creates an ArgumentParser to parse the command line options. """
    help_desc = 'Curator for Elasticsearch indices. See http://github.com/elasticsearch/curator/wiki'
    try:
        import argparse
        parser = argparse.ArgumentParser(description=help_desc)
        parser.add_argument('-v', '--version', action='version', version='%(prog)s '+__version__)
    except ImportError:
        print('{0} requires module argparse.  Try: pip install argparse'.format(sys.argv[0]))
        sys.exit(1)

    # Common args
    parser.add_argument('--host', help='Elasticsearch host. Default: localhost', default=DEFAULT_ARGS['host'])
    parser.add_argument('--url_prefix', help='Elasticsearch http url prefix. Default: none', default=DEFAULT_ARGS['url_prefix'])
    parser.add_argument('--port', help='Elasticsearch port. Default: 9200', default=DEFAULT_ARGS['port'], type=int)
    parser.add_argument('--ssl', help='Connect to Elasticsearch through SSL. Default: false', action='store_true', default=DEFAULT_ARGS['ssl'])
    parser.add_argument('--auth', help='Use Basic Authentication ex: user:pass Default: None', default=DEFAULT_ARGS['auth'])
    parser.add_argument('--timeout', help='Connection timeout in seconds. Default: 30', default=DEFAULT_ARGS['timeout'], type=int)
    parser.add_argument('--master-only', dest='master_only', action='store_true', help='Verify that the node is the elected master before continuing', default=False)
    parser.add_argument('-n', '--dry-run', action='store_true', help='If true, does not perform any changes to the Elasticsearch indices.', default=DEFAULT_ARGS['dry_run'])
    parser.add_argument('-D', '--debug', dest='debug', action='store_true', help='Debug mode', default=DEFAULT_ARGS['debug'])
    parser.add_argument('--loglevel', dest='log_level', action='store', help='Log level', default=DEFAULT_ARGS['log_level'], type=str)
    parser.add_argument('--logfile', dest='log_file', help='log file', type=str)
    parser.add_argument('--logformat', dest='logformat', help='Log output format [default|logstash]. Default: default', default=DEFAULT_ARGS['logformat'], type=str)

    # Command sub_parsers
    subparsers = parser.add_subparsers(
            title='Commands', dest='command', description='Select one of the following commands:',
            help='Run: ' + sys.argv[0] + ' COMMAND --help for command-specific help.')

    # Alias
    parser_alias = subparsers.add_parser('alias', help='Aliasing operations')
    parser_alias.set_defaults(func=curator.alias)
    add_common_args(parser_alias)
    parser_alias.add_argument('--alias', required=True, help='Alias name', type=str)
    alias_group = parser_alias.add_mutually_exclusive_group()
    alias_group.add_argument('--alias-older-than', help='Add indices older than n TIME_UNITs to alias', type=int)
    alias_group.add_argument('--unalias-older-than', help='Remove indices older than n TIME_UNITs from alias', type=int)

    # Allocation
    parser_allocation = subparsers.add_parser('allocation', help='Apply required index routing allocation rule')
    parser_allocation.set_defaults(func=curator.allocation)
    add_common_args(parser_allocation)
    parser_allocation.add_argument('--older-than', required=True, help='Apply rule to indices older than n TIME_UNITs', type=int)
    parser_allocation.add_argument('--rule', required=True, help='Routing allocation rule to apply, e.g. tag=ssd', type=str)

    # Bloom
    parser_bloom = subparsers.add_parser('bloom', help='Disable bloom filter cache for indices')
    parser_bloom.set_defaults(func=curator.bloom)
    add_common_args(parser_bloom)
    parser_bloom.add_argument('--older-than', required=True, help='Disable bloom filter cache for indices older than n TIME_UNITs', type=int)

    # Close
    parser_close = subparsers.add_parser('close', help='Close indices')
    parser_close.set_defaults(func=curator.close)
    add_common_args(parser_close)
    parser_close.add_argument('--older-than', required=True, help='Close indices older than n TIME_UNITs', type=int)

    # Delete
    parser_delete = subparsers.add_parser('delete', help='Delete indices')
    parser_delete.set_defaults(func=curator.delete)
    add_common_args(parser_delete)
    delete_group = parser_delete.add_mutually_exclusive_group()
    delete_group.add_argument('--older-than', help='Delete indices older than n TIME_UNITs', type=int)
    delete_group.add_argument('--disk-space', help='Delete indices beyond DISK_SPACE gigabytes.', type=float)

    # Optimize
    parser_optimize = subparsers.add_parser('optimize', help='Optimize indices')
    parser_optimize.set_defaults(func=curator.optimize)
    add_common_args(parser_optimize)
    parser_optimize.add_argument('--older-than', required=True, help='Optimize indices older than n TIME_UNITs', type=int)
    parser_optimize.add_argument('--max_num_segments', help='Optimize segment count to n segments per shard.', default=DEFAULT_ARGS['max_num_segments'], type=int)
    parser_optimize.add_argument('--delay', help='Number of seconds to delay after optimizing an index.', type=int, default=0)

    # Show indices
    parser_show = subparsers.add_parser('show', help='Show indices or snapshots')
    parser_show.set_defaults(func=show)
    parser_show.add_argument('--prefix', help='Define a prefix. Index name = PREFIX + TIMESTRING + SUFFIX. Default: logstash-', default=DEFAULT_ARGS['prefix'])
    parser_show.add_argument('--suffix', help='Define a suffix. Index name = PREFIX + TIMESTRING + SUFFIX. Default: None', default='')
    parser_show.add_argument('--repository', type=str, help='Repository name (required for --show-repositories)')
    show_group = parser_show.add_mutually_exclusive_group()
    show_group.add_argument('--show-indices', help='Show indices matching PREFIX', action='store_true')
    show_group.add_argument('--show-snapshots', help='Show snapshots in REPOSITORY', action='store_true')
    parser_show.add_argument('--snapshot-prefix', type=str, help='Override default name.', default=DEFAULT_ARGS['snapshot_prefix'])
    parser_show.add_argument('--exclude-pattern', help='Exclude indices matching provided pattern, e.g. 2014.06.08', type=str, default=None)
    

    # Snapshot
    parser_snapshot = subparsers.add_parser('snapshot', help='Take snapshots of indices (Backup)')
    parser_snapshot.set_defaults(func=curator.snapshot)
    add_common_args(parser_snapshot)
    parser_snapshot.add_argument('--repository', required=True, type=str, help='Repository name')
    parser_snapshot.add_argument('--snapshot-name', type=str, help='Override default name.')
    parser_snapshot.add_argument('--snapshot-prefix', type=str, help='Override default name.', default=DEFAULT_ARGS['snapshot_prefix'])

    snapshot_group = parser_snapshot.add_mutually_exclusive_group()
    snapshot_group.add_argument('--older-than', type=int, help='Capture snapshots for indices older than n TIME_UNITs.')
    snapshot_group.add_argument('--all-indices', action='store_true', help='Capture "_all" indices (Elasticsearch default).', default=DEFAULT_ARGS['all_indices'])
    snapshot_group.add_argument('--most-recent', type=int, help='Capture snapshots for n most recent number of indices.')
    snapshot_group.add_argument('--delete-older-than', type=int, help='Delete snapshots older than n TIME_UNITs.')

    parser_snapshot.add_argument('--no_wait_for_completion', action='store_false',
                                help='Do not wait until complete to return. Waits by default.', default=DEFAULT_ARGS['wait_for_completion'])
    parser_snapshot.add_argument('--ignore_unavailable', action='store_true',
                                help='Ignore unavailable shards/indices. Default=False', default=DEFAULT_ARGS['ignore_unavailable'])
    parser_snapshot.add_argument('--include_global_state', action='store_false',
                                help='Store cluster global state with snapshot. Default=True', default=DEFAULT_ARGS['include_global_state'])
    parser_snapshot.add_argument('--partial', action='store_true',
                                help='Do not fail if primary shard is unavailable. Default=False', default=DEFAULT_ARGS['partial'])

    return parser

class Whitelist(logging.Filter):
    def __init__(self, *whitelist):
        self.whitelist = [logging.Filter(name) for name in whitelist]

    def filter(self, record):
        return any(f.filter(record) for f in self.whitelist)

def show(client, **kwargs):
    """
    Show indices or snapshots matching supplied parameters and exit.
    
    :arg client: The Elasticsearch client connection
    :arg data_type: Either ``index`` or ``snapshot``
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg repository: The Elasticsearch snapshot repository to use (only with
        snapshots)
    :arg snapshot_prefix: Override the default with this value. Defaults to
        ``curator-``
    """
    if kwargs['show_indices']:
        for index_name in curator.get_indices(client, prefix=kwargs['prefix'], suffix=kwargs['suffix']):
            print('{0}'.format(index_name))
        sys.exit(0)
    elif kwargs['show_snapshots']:
        for snapshot in curator.get_snaplist(client, kwargs['repository'], snapshot_prefix=kwargs['snapshot_prefix']):
            print('{0}'.format(snapshot))
        sys.exit(0)

def check_version(client):
    """
    Verify version is within acceptable range.  Exit with error if it is not.
    
    :arg client: The Elasticsearch client connection
    """
    version_number = curator.get_version(client)
    logger.debug('Detected Elasticsearch version {0}'.format(".".join(map(str,version_number))))
    if version_number >= version_max or version_number < version_min:
        print('Expected Elasticsearch version range > {0} < {1}'.format(".".join(map(str,version_min)),".".join(map(str,version_max))))
        print('ERROR: Incompatible with version {0} of Elasticsearch.  Exiting.'.format(".".join(map(str,version_number))))
        sys.exit(1)

def validate_timestring(timestring, time_unit):
    """
    Validate that the appropriate element(s) for time_unit are in the timestring.
    e.g. If "weeks", we should see %U or %W, if hours %H, etc.
    
    Exit with error on failure.
    
    :arg timestring: An strftime string to match the datestamp in an index name.
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``.
    """
    fail = True
    if time_unit == 'hours':
        if '%H' in timestring:
            fail = False
    elif time_unit == 'days':
        if '%d' in timestring:
            fail = False
    elif time_unit == 'weeks':
        if '%W' in timestring:
            fail = False
        elif '%U' in timestring:
            fail = False
    elif time_unit == 'months':
        if '%m' in timestring:
            fail = False
    if fail:
        print('Timestring {0} does not match time unit {1}'.format(timestring, time_unit))
        sys.exit(1)
    return
    
def main():
    start = time.time()

    parser = make_parser()
    arguments = parser.parse_args()

    # Initialize timeout_override
    timeout_override = True if arguments.command == 'optimize' else False

    # Argparse nearly gets all conditions covered.
    # These remain because mutually exclusive arguments must be optional.
    if arguments.command == 'alias':
        if not arguments.alias_older_than and not arguments.unalias_older_than:
            print('{0} delete: error: expect one of --alias-older-than or --unalias-older-than'.format(sys.argv[0]))
            sys.exit(1)

    if arguments.command == 'delete':
        if not arguments.older_than and not arguments.disk_space:
            print('{0} delete: error: expect one of --older-than or --disk-space'.format(sys.argv[0]))
            sys.exit(1)

    if arguments.command == 'show':
        # Do not log and force dry-run if we opt to show indices or snapshots.
        arguments.log_file = os.devnull
        arguments.dry_run = True
        if not arguments.show_indices and not arguments.show_snapshots:
            print('{0} show: error: expect one of --show-indices or --show-snapshots'.format(sys.argv[0]))
            sys.exit(1)
        if arguments.show_snapshots and not arguments.repository:
            print('{0} show: error: --repository required with --show-snapshots'.format(sys.argv[0]))
            sys.exit(1)

    if arguments.command == 'snapshot':
        if not arguments.older_than and not arguments.most_recent and not arguments.delete_older_than and not arguments.all_indices:
            print('{0} snapshot: error: expect one of --all-indices, --older-than, --most-recent, or --delete-older-than'.format(sys.argv[0]))
            sys.exit(1)
        if arguments.older_than or arguments.most_recent:
            timeout_override = True

    # Setup logging
    if arguments.debug:
        numeric_log_level = logging.DEBUG
        format_string = '%(asctime)s %(levelname)-9s %(name)22s %(funcName)22s:%(lineno)-4d %(message)s'
    else:
        numeric_log_level = getattr(logging, arguments.log_level.upper(), None)
        format_string = '%(asctime)s %(levelname)-9s %(message)s'
        if not isinstance(numeric_log_level, int):
            raise ValueError('Invalid log level: %s' % arguments.log_level)
    
    date_string = None
    if arguments.logformat == 'logstash':
        os.environ['TZ'] = 'UTC'
        time.tzset()
        format_string = '{"@timestamp":"%(asctime)s.%(msecs)03dZ", "loglevel":"%(levelname)s", "name":"%(name)s", "function":"%(funcName)s", "linenum":"%(lineno)d", "message":"%(message)s"}'
        date_string = '%Y-%m-%dT%H:%M:%S'

    logging.basicConfig(level=numeric_log_level,
                        format=format_string,
                        datefmt=date_string,
                        stream=open(arguments.log_file, 'a') if arguments.log_file else sys.stderr)

    # Filter out logging from Elasticsearch and associated modules by default
    if not arguments.debug:
        for handler in logging.root.handlers:
            handler.addFilter(Whitelist('root', '__main__', 'curator', 'curator.curator'))

    # Setting up NullHandler to handle nested elasticsearch.trace Logger instance in elasticsearch python client
    logging.getLogger('elasticsearch.trace').addHandler(NullHandler())

    logging.info("Job starting...")

    if arguments.dry_run:
        logging.info("DRY RUN MODE.  No changes will be made.")

    # Override the timestamp in case the end-user doesn't.
    if timeout_override and arguments.timeout == 30:
        logger.info('Default timeout of 30 seconds is too low for command {0}.  Overriding to 21,600 seconds (6 hours).'.format(arguments.command.upper()))
        arguments.timeout = 21600

    client = elasticsearch.Elasticsearch(host=arguments.host, http_auth=arguments.auth, port=arguments.port, url_prefix=arguments.url_prefix, timeout=arguments.timeout, use_ssl=arguments.ssl)
    
    # Verify the version is acceptable.
    check_version(client)
    
    if arguments.master_only and not curator.is_master_node(client):
        logger.info('Master-only flag detected. Connected to non-master node. Aborting.')
        sys.exit(0)

    if arguments.command != "show":
        if arguments.timestring:
            validate_timestring(arguments.timestring, arguments.time_unit)
        else: # Set default timestrings
            arguments.timestring = DATEMAP[arguments.time_unit]
            logging.debug("Setting default timestring for {0} to {1}".format(arguments.time_unit, arguments.timestring))
        logging.debug("Matching indices with pattern: {0}{1}".format(arguments.prefix,arguments.timestring))

    # Execute the command specified in the arguments
    argdict = arguments.__dict__
    logging.debug("argdict = {0}".format(argdict))
    arguments.func(client, **argdict)

    logger.info('Done in {0}.'.format(timedelta(seconds=time.time()-start)))

if __name__ == '__main__':
    main()
    



import time
import logging
import re
from datetime import timedelta, datetime, date

import elasticsearch

__version__ = '2.0.0'

logger = logging.getLogger(__name__)

DATE_REGEX = {
    'Y' : '4',
    'm' : '2',
    'W' : '2',
    'U' : '2',
    'd' : '2',
    'H' : '2',
}

# Information retrieval
## Date & Time
### Date Regex
def get_date_regex(timestring):
    """
    Return a regex string based on a provided strftime timestring.
    
    :arg timestring: An strftime pattern
    :rtype: Regex as string
    """
    prev = ''; curr = ''; regex = ''
    for s in range(0, len(timestring)):
        curr = timestring[s]
        if curr == '%':
            pass
        elif curr in DATE_REGEX and prev == '%':
            regex += '\d{' + DATE_REGEX[curr] + '}'
        else:
            regex += "\\" + curr
        prev = curr
    return regex

### Index time
def get_index_time(index_timestamp, timestring):
    """
    Return the datetime extracted from the index name, which is the index
    creation time.

    :arg index_timestamp: The timestamp extracted from an index name
    :arg timestring: An strftime pattern
    :rtype: Datetime object
    """
    # Compensate for week of year by appending '%w' to the timestring
    # and '1' (Monday) to index_timestamp
    if '%W' in timestring:
        timestring += '%w'
        index_timestamp += '1'
    elif '%U' in timestring:
        timestring += '%w'
        index_timestamp += '1'
    elif '%m' in timestring:
        if not '%d' in timestring:
            timestring += '%d'
            index_timestamp += '1'
    return datetime.strptime(index_timestamp, timestring)

### Index time if unit = months
def get_target_month(month_count, utc_now=None):
    """
    Return datetime object for # of *full* months older than ``month_count`` from
    now, or ``utc_now``, if provided.
    
    :arg month_count: Number of *full* months
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    :rtype: Datetime object
    """
    utc_now = date(utc_now.year, utc_now.month, 1) if utc_now else date.today()
    target_date = date(utc_now.year, utc_now.month, 1)

    for i in range(0, month_count):
        if target_date.month == 1:
            target_date = date(target_date.year-1, 12, 1)
        else:
            target_date = date(target_date.year, target_date.month-1, 1)

    return datetime(target_date.year, target_date.month, target_date.day)

### Cutoff time
def get_cutoff(older_than=999999, time_unit='days', utc_now=None):
    """
    Find the cutoff time based on ``older_than`` and ``time_unit``.

    :arg older_than: ``time_unit`` multiplier
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    :rtype: Datetime object
    """
    # time-injection for test purposes only
    utc_now = utc_now if utc_now else datetime.utcnow()
    # reset to start of the period to be sure we are not retiring a human by mistake
    utc_now = utc_now.replace(minute=0, second=0, microsecond=0)

    if time_unit == 'days':
        utc_now = utc_now.replace(hour=0)
    if time_unit == 'weeks':
        # Since week math always uses Monday as the start of the week,
        # this work-around resets utc_now to be Monday of the current week.
        weeknow = utc_now.strftime('%Y-%W')
        utc_now = get_index_time(weeknow, '%Y-%W')
    if time_unit == 'months':
        utc_now = utc_now.replace(hour=0)
        cutoff = get_target_month(older_than, utc_now=utc_now)
    else:
        # This cutoff must be a multiple of time_units
        cutoff = utc_now - timedelta(**{time_unit: (older_than - 1)})
    return cutoff

## Alias information
def get_alias(client, alias):
    """
    Return information about the specified alias.

    :arg client: The Elasticsearch client connection
    :arg alias: Alias name to operate on.
    :rtype: list of strings
    """
    if client.indices.exists_alias(alias):
        return client.indices.get_alias(name=alias).keys()
    else:
        logger.error('Unable to find alias {0}.'.format(alias))
        return False

## Is the index closed?
def index_closed(client, index_name):
    """
    Return `True` if the indicated index is closed.

    :arg client: The Elasticsearch client connection
    :arg index_name: The index name
    :rtype: bool
    """
    index_metadata = client.cluster.state(
        index=index_name,
        metric='metadata',
    )
    return index_metadata['metadata']['indices'][index_name]['state'] == 'close'

## Get matching indices
def get_indices(client, prefix='logstash-', suffix='', exclude_pattern=None):
    """
    Return a sorted list of indices matching ``prefix`` and ``suffix`` and
    optionally filtered by ``exclude_pattern``.
    
    :arg client: The Elasticsearch client connection
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg exclude_pattern: Exclude indices matching the provided regular
        expression.
    :rtype: list of strings
    """
    _indices = sorted(client.indices.get_settings(index='*', params={'expand_wildcards': 'open,closed'}).keys())
    if prefix:
        prefix = '.' + prefix if prefix[0] == '*' else prefix
    if suffix:
        suffix = '.' + suffix if suffix[0] == '*' else suffix
    regex = "^" + prefix + ".*" + suffix + "$"
    _fixes = re.compile(regex)
    _indices = list(filter(lambda x: _fixes.search(x), _indices))
    if '.marvel-kibana' in _indices:
        _indices.remove('.marvel-kibana')
    if 'kibana-int' in _indices:
        _indices.remove('kibana-int')
    if exclude_pattern:
        pattern = re.compile(exclude_pattern)
        return list(filter(lambda x: not pattern.search(x), _indices))
    else:
        return _indices

## Snapshot information
### List of matching snapshots
def get_snaplist(client, repository='', snapshot_prefix='curator-'):
    """
    Get ``_all`` snapshots containing ``snapshot_prefix`` from repository and
    return a list.
    
    :arg client: The Elasticsearch client connection
    :arg repository: The Elasticsearch snapshot repository to use
    :arg snapshot_prefix: Override the default with this value. Defaults to
        ``curator-``
    :rtype: list of strings
    """
    retval = []
    try:
        allsnaps = client.snapshot.get(repository=repository, snapshot="_all")['snapshots']
        snaps = [snap['snapshot'] for snap in allsnaps if 'snapshot' in snap.keys()]
        if snapshot_prefix:
            snapshot_prefix = '.' + snapshot_prefix if snapshot_prefix[0] == '*' else snapshot_prefix
        regex = "^" + snapshot_prefix + ".*" + "$"
        pattern = re.compile(regex)
        return list(filter(lambda x: pattern.search(x), snaps))
    except elasticsearch.NotFoundError as e:
        logger.error("Error: {0}".format(e))
    return retval

### Repository information
def get_repository(client, repository=''):
    """
    Return configuration information for the indicated repository.

    :arg client: The Elasticsearch client connection
    :arg repository: The Elasticsearch snapshot repository to use
    :rtype: dict
    """
    try:
        return client.snapshot.get_repository(repository=repository)
    except elasticsearch.NotFoundError as e:
        logger.info("Repository {0} not found.  Error: {1}".format(repository, e))
        return None

### Single snapshot information
def get_snapshot(client, repository='', snapshot=''):
    """
    Return information about a snapshot (or a comma-separated list of snapshots)

    :arg client: The Elasticsearch client connection
    :arg repository: The Elasticsearch snapshot repository to use
    :arg snapshot: The snapshot name, or a comma-separated list of snapshots
    :rtype: dict
    """
    try:
        return client.snapshot.get(repository=repository, snapshot=snapshot)
    except elasticsearch.NotFoundError as e:
        logger.info("Snapshot or repository {0} not found.  Error: {1}".format(snapshot, e))
        return None

## ES version
def get_version(client):
    """
    Return the ES version number as a tuple.
    
    :arg client: The Elasticsearch client connection
    :rtype: tuple
    """
    version = client.info()['version']['number']
    return tuple(map(int, version.split('.')))

## Segment count
def get_segmentcount(client, index_name):
    """
    Return a tuple of ``(shardcount, segmentcount)`` from the provided
    ``index_name``.

    :arg client: The Elasticsearch client connection
    :arg index_name: The index name
    :rtype: tuple
    """
    shards = client.indices.segments(index=index_name)['indices'][index_name]['shards']
    segmentcount = 0
    totalshards = 0 # We will increment this manually to capture all replicas...
    for shardnum in shards:
        for shard in range(0,len(shards[shardnum])):
            segmentcount += shards[shardnum][shard]['num_search_segments']
            totalshards += 1
    return totalshards, segmentcount

## Is the current node the elected master?
def is_master_node(client):
    """
    Return `True` if the connected client node is the elected master node in
    the Elasticsearch cluster, otherwise return `False`.
    
    :arg client: The Elasticsearch client connection
    :rtype: bool
    """
    my_node_id = client.nodes.info('_local')['nodes'].keys()[0]
    master_node_id = client.cluster.state(metric='master_node')['master_node']
    return my_node_id == master_node_id

## Matching indices or snapshots
def get_object_list(client, data_type='index', prefix='logstash-', suffix='', repository=None, snapshot_prefix='curator-', exclude_pattern=None, **kwargs):
    """
    Return a list of indices matching ``prefix`` and ``suffix`` or snapshots
    matching ``snapshot_prefix``.
    
    Can optionally exclude by regular expression pattern with
    ``exclude_pattern``.
    
    :arg client: The Elasticsearch client connection
    :arg data_type: Either ``index`` or ``snapshot``
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg repository: The Elasticsearch snapshot repository to use (only with
        snapshots)
    :arg snapshot_prefix: Override the default with this value. Defaults to
        ``curator-``
    :arg exclude_pattern: Exclude indices matching the provided regular
        expression.
    :rtype: list of strings
    """
    if data_type == 'index':
        object_list = get_indices(client, prefix=prefix, suffix=suffix)
    elif data_type == 'snapshot':
        if repository:
            object_list = get_snaplist(client, repository=repository, snapshot_prefix=snashot_prefix)
        else:
            logger.error('Repository name not specified. Returning empty list.')
            object_list = []
    else:
        object_list = []
        logger.error('data_type \'{0}\' is neither \'index\' nor \'snapshot\'.  Returning empty list.'.format(data_type))
    if exclude_pattern:
        pattern = re.compile(exclude_pattern)
        return list(filter(lambda x: not pattern.search(x), object_list))
    else:
        return object_list

# Filtering
## By timestamp
def filter_by_timestamp(object_list=[], timestring=None, time_unit='days',
                        older_than=999999, prefix='logstash-', suffix='',
                        snapshot_prefix='curator-', utc_now=None, **kwargs):
    """
    Pass in a list of indices or snapshots. Return a list of objects older
    than *n* ``time_unit``\s matching ``prefix``, ``timestring``, and
    ``suffix``.
    
    :arg object_list: A list of indices or snapshots
    :arg timestring: An strftime string to match the datestamp in an index name.
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``
    :arg older_than: Indices older than the indicated number of whole
        ``time_units`` will be operated on.
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg snapshot_prefix: Override the default with this value. Defaults to
        ``curator-``
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    :rtype: generator object (list of strings)
    """
    object_type = kwargs['object_type'] if 'object_type' in kwargs else 'index'
    if prefix:
        prefix = '.' + prefix if prefix[0] == '*' else prefix
    if suffix:
        suffix = '.' + suffix if suffix[0] == '*' else suffix
    if snapshot_prefix:
        snapshot_prefix = '.' + snapshot_prefix if snapshot_prefix[0] == '*' else snapshot_prefix
    dateregex = get_date_regex(timestring)
    if object_type == 'index':
        regex = "^" + prefix + "(" + dateregex + ")" + suffix + "$"
    elif object_type == 'snapshot':
        regex = "(" + "^" + snapshot_prefix + '.*' + ")"

    cutoff = get_cutoff(older_than=older_than, time_unit=time_unit, utc_now=utc_now)
    
    for object_name in object_list:
        retval = object_name
        if object_type == 'index':
            try:
                index_timestamp = re.search(regex, object_name).group(1)
            except AttributeError as e:
                logger.debug('Unable to match {0} with regular expression {1}.  Error: {2}'.format(object_name, regex, e))
                continue
            try:
                object_time = get_index_time(index_timestamp, timestring)
            except ValueError:
                logger.error('Could not find a valid timestamp for {0} with timestring {1}'.format(object_name, timestring))
                continue
        elif object_type == 'snapshot':
            try:
                retval = re.search(regex, object_name['snapshot']).group(1)
            except AttributeError as e:
                logger.debug('Unable to match {0} with regular expression {1}.  Error: {2}'.format(retval, regex, e))
                continue
            try:
                object_time = datetime.utcfromtimestamp(object_name['start_time_in_millis']/1000.0)
            except AttributeError as e:
                logger.debug('Unable to compare time from snapshot {0}.  Error: {1}'.format(object_name, e))
                continue
            # if the index is older than the cutoff
        if object_time < cutoff:
            yield retval
        else:
            logger.info('{0} is within the threshold period ({1} {2}).'.format(retval, older_than, time_unit))

## By space
def filter_by_space(client, disk_space=2097152.0, prefix='logstash-', suffix='',
                    exclude_pattern=None, **kwargs):
    """
    Yield a list of indices to delete based on space consumed, starting with
    the oldest.
    
    :arg client: The Elasticsearch client connection
    :arg disk_space: Delete indices over *n* gigabytes, starting from the
        oldest indices.
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg exclude_pattern: Exclude indices matching the provided regular
        expression.
    :rtype: generator object (list of strings)
    """

    disk_usage = 0.0
    disk_limit = disk_space * 2**30

    # Use of exclude_pattern here could be _very_ important if you don't
    # want an index pruned even if it is old.
    exclude_pattern = kwargs['exclude_pattern'] if 'exclude_pattern' in kwargs else ''

    # These two lines allow us to use common filtering by regex before
    # gathering stats.  However, there are still pitfalls.  You may still
    # wind up deleting more of one kind of index than another if you have
    # multiple kinds.  Also, it still won't work on closed indices, so we
    # must filter them out.
    all_indices = get_indices(client, prefix=prefix, suffix=suffix, exclude_pattern=exclude_pattern)
    not_closed = [i for i in all_indices if not index_closed(client, i)]
    csv_indices = ','.join(not_closed)

    stats = client.indices.status(index=csv_indices)
    
    sorted_indices = sorted(
        (
            (index_name, index_stats['index']['primary_size_in_bytes'])
            for (index_name, index_stats) in stats['indices'].items()
        ),
        reverse=True
    )

    for index_name, index_size in sorted_indices:
        disk_usage += index_size

        if disk_usage > disk_limit:
            yield index_name
        else:
            logger.info('skipping {0}, summed disk usage is {1:.3f} GB and disk limit is {2:.3f} GB.'.format(index_name, disk_usage/2**30, disk_limit/2**30))

# Operations
## Single-index operations
### Alias
#### Add to alias
def add_to_alias(client, index_name, alias=None, **kwargs):
    """
    Add indicated index to the specified alias.

    :arg client: The Elasticsearch client connection
    :arg index_name: The index name
    :arg alias: Alias name to operate on.
    :rtype: bool
    """
    if not alias: # This prevents _all from being aliased by accident...
        logger.error('No alias provided.')
        return True
    if not client.indices.exists_alias(alias):
        logger.error('Skipping index {0}: Alias {1} does not exist.'.format(index_name, alias))
        return True
    else:
        indices_in_alias = client.indices.get_alias(alias)
        if not index_name in indices_in_alias:
            client.indices.update_aliases(body={'actions': [{ 'add': { 'index': index_name, 'alias': alias}}]})
        else:
            logger.info('Skipping index {0}: Index already exists in alias {1}...'.format(index_name, alias))
            return True

#### Remove from alias
def remove_from_alias(client, index_name, alias=None, **kwargs):
    """
    Remove the indicated index from the specified alias.

    :arg client: The Elasticsearch client connection
    :arg index_name: The index name
    :arg alias: Alias name to operate on.
    :rtype: bool
    """
    indices_in_alias = get_alias(client, alias)
    if not indices_in_alias:
        return True
    if index_name in indices_in_alias:
        client.indices.update_aliases(body={'actions': [{ 'remove': { 'index': index_name, 'alias': alias}}]})
    else:
        logger.info('Index {0} does not exist in alias {1}; skipping.'.format(index_name, alias))
        return True

### Allocation
def apply_allocation_rule(client, index_name, rule=None, **kwargs):
    """
    Apply a required allocation rule to an index.  See:
    http://www.elasticsearch.org/guide/en/elasticsearch/reference/current/index-modules-allocation.html#index-modules-allocation

    :arg client: The Elasticsearch client connection
    :arg index_name: The index name
    :arg rule: The routing allocation rule to apply, e.g. ``tag=ssd``.  Must be
        in the format of ``key=value``, and should match values declared on the
        correlating nodes in your cluster.
    """
    if not rule:
      logger.error('No rule provided for {0}.'.format(index_name))
      return True
    key = rule.split('=')[0]
    value = rule.split('=')[1]
    if index_closed(client, index_name):
      logger.info('Skipping index {0}: Already closed.'.format(index_name))
      return True
    else:
      logger.info('Updating index setting index.routing.allocation.require.{0}={1}'.format(key,value))
      client.indices.put_settings(index=index_name, body='index.routing.allocation.require.{0}={1}'.format(key,value))

### Bloom
def disable_bloom_filter(client, index_name, **kwargs):
    """
    Disable the bloom filter cache for the specified index.

    :arg client: The Elasticsearch client connection
    :arg index_name: The index name
    """
    if index_closed(client, index_name): # Don't try to disable bloom filter on a closed index.  It will re-open them
        logger.info('Skipping index {0}: Already closed.'.format(index_name))
        return True
    else:
        client.indices.put_settings(index=index_name, body='index.codec.bloom.load=false')

### Close
def close_index(client, index_name, **kwargs):
    """
    Close the indicated index.  Flush before closing.
    
    :arg client: The Elasticsearch client connection
    :arg index_name: The index name
    """
    if index_closed(client, index_name):
        logger.info('Skipping index {0}: Already closed.'.format(index_name))
        return True
    else:
        client.indices.flush(index=index_name)
        client.indices.close(index=index_name)

### Delete
def delete_index(client, index_name, **kwargs):
    """
    Delete the indicated index.
    
    :arg client: The Elasticsearch client connection
    :arg index_name: The index name
    """
    client.indices.delete(index=index_name)

### Optimize
def optimize_index(client, index_name, max_num_segments=2, **kwargs):
    """
    Optimize (Lucene forceMerge) index to ``max_num_segments`` per shard
    
    :arg client: The Elasticsearch client connection
    :arg index_name: The index name
    :arg max_num_segments: Merge to this number of segments per shard.
    """
    if index_closed(client, index_name): # Don't try to optimize a closed index
        logger.info('Skipping index {0}: Already closed.'.format(index_name))
        return True
    else:
        shards, segmentcount = get_segmentcount(client, index_name)
        logger.debug('Index {0} has {1} shards and {2} segments total.'.format(index_name, shards, segmentcount))
        if segmentcount > (shards * max_num_segments):
            logger.info('Optimizing index {0} to {1} segments per shard.  Please wait...'.format(index_name, max_num_segments))
            client.indices.optimize(index=index_name, max_num_segments=max_num_segments)
        else:
            logger.info('Skipping index {0}: Already optimized.'.format(index_name))
            return True

## Snapshots
### Create a snapshot body
def create_snapshot_body(indices, ignore_unavailable=False,
                         include_global_state=True, partial=False, **kwargs):
    """
    Create the request body for creating a snapshot from the provided
    arguments.

    :arg indices: A single index, or list of indices to snapshot.
    :arg ignore_unavailable: Boolean. Ignore unavailable shards/indices.
        Default is `False`
    :arg include_global_state: Boolean. Store cluster global state with snapshot.
        Default is `True`
    :arg partial: Boolean. Do not fail if primary shard is unavailable. Default
        is `False`
    """
    body = {
        "ignore_unavailable": ignore_unavailable,
        "include_global_state": include_global_state,
        "partial": partial,
    }
    if indices == '_all':
        body["indices"] = indices
    else:
        if type(indices) is not type(list()):   # in case of a single value passed
            indices = [indices]
        body["indices"] = ','.join(sorted(indices))
    return body

### Create a snapshot
def create_snapshot(client, indices='_all', snapshot_name=None,
                    snapshot_prefix='curator-', repository='',
                    ignore_unavailable=False, include_global_state=True,
                    partial=False, wait_for_completion=True, **kwargs):
    """
    Create a snapshot of provided indices (or ``_all``).

    :arg client: The Elasticsearch client connection
    :arg indices: A list of indices to snapshot. Default is ``_all``
    :arg snapshot_name: What to name the snapshot. ``snapshot_prefix`` +
        datestamp if omitted.
    :arg snapshot_prefix: Override the default with this value. Defaults to
        ``curator-``
    :arg repository: The Elasticsearch snapshot repository to use
    :arg wait_for_completion: Wait (or not) for the operation
        to complete before returning.  Waits by default, i.e. Default is
        `True`
    :type wait_for_completion: bool
    :arg ignore_unavailable: Ignore unavailable shards/indices.
        Default is `False`
    :type ignore_unavailable: bool
    :arg include_global_state: Store cluster global state with snapshot.
        Default is `True`
    :type include_global_state: bool
    :arg partial: Do not fail if primary shard is unavailable. Default
        is `False`
    :type partial: bool
    """
    # Return True if it is skipped
    if not repository:
        logger.error("Unable to create snapshot. Repository name not provided.")
        return True
    try:
        if not indices == '_all':
            if type(indices) == type(list()):
                indices = [i for i in indices if not index_closed(client, i)]
            else:
                indices = indices if not index_closed(client, indices) else ''
        body=create_snapshot_body(indices, ignore_unavailable=ignore_unavailable, include_global_state=include_global_state, partial=partial)
        datestamp = datetime.utcnow().strftime('%Y%m%d%H%M%S')
        snapshot_name = snapshot_name if snapshot_name else snapshot_prefix + datestamp
        all_snaps = get_snaplist(client, repository=repository, snapshot_prefix=snapshot_prefix)
        if not snapshot_name in all_snaps and len(indices) > 0:
            client.snapshot.create(repository=repository, snapshot=snapshot_name, body=body, wait_for_completion=wait_for_completion)
        elif len(indices) == 0:
            logger.warn("No indices provided.")
            return True
        else:
            logger.info("Skipping: A snapshot with name '{0}' already exists.".format(snapshot_name))
            return True
    except elasticsearch.RequestError as e:
        logger.error("Unable to create snapshot {0}.  Error: {1} Check logs for more information.".format(snapshot_name, e))
        return True

### Delete a snapshot
def delete_snapshot(client, snapshot_name, repository=None, **kwargs):
    """
    Delete a snapshot (or comma-separated list of snapshots)

    :arg client: The Elasticsearch client connection
    :arg snapshot_name: The snapshot name
    :arg repository: The Elasticsearch snapshot repository to use
    """
    try:
        client.snapshot.delete(repository=repository, snapshot=snapshot_name)
    except elasticsearch.RequestError as e:
        logger.error("Unable to delete snapshot {0}.  Error: {1} Check logs for more information.".format(snapshot_name, e))

# Operations typically used by the curator_script, directly or indirectly
## Loop through a list of objects and perform the indicated operation
def _op_loop(client, object_list, op=None, dry_run=False, **kwargs):
    """
    Perform the ``op`` on indices or snapshots in the ``object_list``
    
    .. note::
       All kwargs are passed on to the specified ``op``.
       
       Any arg the ``op`` needs *must* be passed in **kwargs**.
    
    :arg client: The Elasticsearch client connection
    :arg object_list: The list of indices or snapshots
    :arg op: The operation to perform on each object in ``object_list``
    :arg dry_run: If true, simulate, but do not perform the operation
    :arg delay: Can be specified to pause after each iteration in the loop.
        Useful to allow cluster to quiesce after heavy I/O optimizations.
    """
    if not op:
        logger.error("No operation specified.")
        return
    prepend = kwargs['prepend'] if 'prepend' in kwargs else ''
    for item in object_list:
        if dry_run: # Don't act on a dry run
            logger.info(prepend + "of {0} operation on {1}".format(op.__name__, item))
            continue
        skipped = op(client, item, **kwargs)
        if skipped:
            continue
        if 'delay' in kwargs:
            if kwargs['delay'] > 0:
                logger.info('Pausing for {0} seconds to allow cluster to quiesce...'.format(kwargs['delay']))
                time.sleep(kwargs['delay'])
        logger.info("{0} operation succeeded on {1}".format(op.__name__, item))

## curator alias [ARGS]
def alias(client, dry_run=False, **kwargs):
    """
    Multiple use cases:
    
    1. Add indices ``alias_older_than`` *n* ``time_unit``\s, matching the given ``timestring``, ``prefix``, and ``suffix`` to ``alias``.
    2. Remove indices ``unalias_older_than`` *n* ``time_unit``\s, matching the given ``timestring``, ``prefix``, and ``suffix`` to ``alias``.

    .. note::
       As this is an iterative function, default values are handled by the
       target function(s).
       
       Unless passed in `kwargs`, parameters other than ``client`` and
       ``dry_run`` will have default values assigned by the functions being
       called:
    
       :py:func:`curator.curator.get_object_list`
    
       :py:func:`curator.curator.filter_by_timestamp`
    
       :py:func:`curator.curator.add_to_alias`

       :py:func:`curator.curator.add_to_alias`
    
    These defaults are included here for documentation.
    
    :arg client: The Elasticsearch client connection
    :arg dry_run: If true, simulate, but do not perform the operation
    :arg alias: Alias name to operate on.
    :arg alias_older_than: Indices older than the indicated number of whole
        ``time_units`` will be operated on.
    :arg unalias_older_than: Indices older than the indicated number of whole
        ``time_units`` will be operated on.
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``.
    :arg timestring: An strftime string to match the datestamp in an index name.
    :arg prefix: A string that comes before the datestamp in an index name. Can
        be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name. Can
        be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg exclude_pattern: Exclude indices matching the provided regular expression.
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    """
    kwargs['prepend'] = "DRY RUN: " if dry_run else ''
    logging.info(kwargs['prepend'] + "Performing alias operations...")
    if kwargs['alias_older_than']:
        kwargs['older_than'] = kwargs['alias_older_than']
        op = add_to_alias
    elif kwargs['unalias_older_than']:
        kwargs['older_than'] = kwargs['unalias_older_than']
        op = remove_from_alias
    else:
        op = None
    index_list = get_object_list(client, **kwargs)
    matching_indices = filter_by_timestamp(object_list=index_list, **kwargs)
    _op_loop(client, matching_indices, op=op, dry_run=dry_run, **kwargs)
    logger.info(kwargs['prepend'] + 'Alias operations completed for specified indices.')

## curator allocation [ARGS]
def allocation(client, dry_run=False, **kwargs):
    """
    Change shard/routing allocation for indices ``older_than`` *n*
    ``time_unit``\s, matching the given ``timestring``, ``prefix``, and
    ``suffix`` by updating/applying the provided ``rule``.
    
    .. note::
       As this is an iterative function, default values are handled by the
       target function(s).
       
       Unless passed in `kwargs`, parameters other than ``client`` and
       ``dry_run`` will have default values assigned by the functions being
       called:
    
       :py:func:`curator.curator.get_object_list`
    
       :py:func:`curator.curator.filter_by_timestamp`
    
       :py:func:`curator.curator.apply_allocation_rule`
    
       These defaults are included here for documentation.

    :arg client: The Elasticsearch client connection
    :arg dry_run: If true, simulate, but do not perform the operation
    :arg older_than: Indices older than the indicated number of whole
        ``time_units`` will be operated on.
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``.
    :arg timestring: An strftime string to match the datestamp in an index name.
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg rule: The routing allocation rule to apply, e.g. ``tag=ssd``.  Must be
        in the format of ``key=value``, and should match values declared on the
        correlating nodes in your cluster.
    :arg exclude_pattern: Exclude indices matching the provided regular
        expression.
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    """
    kwargs['prepend'] = "DRY RUN: " if dry_run else ''
    logging.info(kwargs['prepend'] + "Applying allocation/routing tags to indices...")
    index_list = get_object_list(client, **kwargs)
    matching_indices = filter_by_timestamp(object_list=index_list, **kwargs)
    _op_loop(client, matching_indices, op=apply_allocation_rule, dry_run=dry_run, **kwargs)
    logger.info(kwargs['prepend'] + 'Allocation/routing tags applied to specified indices.')

## curator bloom [ARGS]
def bloom(client, dry_run=False, **kwargs):
    """
    Disable bloom filter cache for indices ``older_than`` *n* ``time_unit``\s,
    matching the given ``timestring``, ``prefix``, and ``suffix``.
    
    .. note::
       As this is an iterative function, default values are handled by the
       target function(s).
       
       Unless passed in `kwargs`, parameters other than ``client`` and
       ``dry_run`` will have default values assigned by the functions being
       called:
    
       :py:func:`curator.curator.get_object_list`
    
       :py:func:`curator.curator.filter_by_timestamp`
    
       :py:func:`curator.curator.disable_bloom_filter`
    
       These defaults are included here for documentation.

    :arg client: The Elasticsearch client connection
    :arg dry_run: If true, simulate, but do not perform the operation
    :arg older_than: Indices older than the indicated number of whole
        ``time_units`` will be operated on.
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``.
    :arg timestring: An strftime string to match the datestamp in an index name.
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg exclude_pattern: Exclude indices matching the provided regular
        expression.
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    """
    kwargs['prepend'] = "DRY RUN: " if dry_run else ''
    logging.info(kwargs['prepend'] + "Disabling the bloom filter cache for indices...")
    index_list = get_object_list(client, **kwargs)
    matching_indices = filter_by_timestamp(object_list=index_list, **kwargs)
    _op_loop(client, matching_indices, op=disable_bloom_filter, dry_run=dry_run, **kwargs)
    logger.info(kwargs['prepend'] + 'Disabled bloom filter cache for specified indices.')

## curator close [ARGS]
def close(client, dry_run=False, **kwargs):
    """
    Close indices ``older_than`` *n* ``time_unit``\s, matching the given
    ``timestring``, ``prefix``, and ``suffix``.
    
    .. note::
       As this is an iterative function, default values are handled by the
       target function(s).
       
       Unless passed in `kwargs`, parameters other than ``client`` and
       ``dry_run`` will have default values assigned by the functions being
       called:
    
       :py:func:`curator.curator.get_object_list`
    
       :py:func:`curator.curator.filter_by_timestamp`
    
       :py:func:`curator.curator.close_index`
    
       These defaults are included here for documentation.

    :arg client: The Elasticsearch client connection
    :arg dry_run: If true, simulate, but do not perform the operation
    :arg older_than: Indices older than the indicated number of whole
        ``time_units`` will be operated on.
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``.
    :arg timestring: An strftime string to match the datestamp in an index name.
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg exclude_pattern: Exclude indices matching the provided regular
        expression.
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    """
    kwargs['prepend'] = "DRY RUN: " if dry_run else ''
    logging.info(kwargs['prepend'] + "Closing indices...")
    index_list = get_object_list(client, **kwargs)
    matching_indices = filter_by_timestamp(object_list=index_list, **kwargs)
    _op_loop(client, matching_indices, op=close_index, dry_run=dry_run, **kwargs)
    logger.info(kwargs['prepend'] + 'Closed specified indices.')

## curator delete [ARGS]
def delete(client, dry_run=False, **kwargs):
    """
    Two use cases for deleting indices:
    
    1. Delete indices ``older_than`` *n* ``time_unit``\s, matching the given ``timestring``, ``prefix``, and ``suffix``.
    2. Delete indices in excess of ``disk_space`` gigabytes if the ``disk_space`` kwarg is present, beginning from the oldest.  Indices must still match ``prefix`` and ``suffix``.  This amount spans all nodes and shards and must be calculated accordingly.  This is not a recommended use-case.

    .. note::
       As this is an iterative function, default values are handled by the
       target function(s).
       
       Unless passed in `kwargs`, parameters other than ``client`` and
       ``dry_run`` will have default values assigned by the functions being
       called:

       :py:func:`curator.curator.get_object_list`

       :py:func:`curator.curator.filter_by_space`

       :py:func:`curator.curator.filter_by_timestamp`

       :py:func:`curator.curator.delete_index`

       These defaults are included here for documentation.

    :arg client: The Elasticsearch client connection
    :arg dry_run: If true, simulate, but do not perform the operation
    :arg older_than: Indices older than the indicated number of whole
        ``time_units`` will be operated on.
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``.
    :arg timestring: An strftime string to match the datestamp in an index name.
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg disk_space: Delete indices over *n* gigabytes, starting from the
        oldest indices.
    :arg exclude_pattern: Exclude indices matching the provided regular
        expression.
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    """
    kwargs['prepend'] = "DRY RUN: " if dry_run else ''
    logging.info(kwargs['prepend'] + "Deleting indices...")
    by_space = kwargs['disk_space'] if 'disk_space' in kwargs else False
    if by_space:
        logger.info(kwargs['prepend'] + 'Deleting by space rather than time.')
        matching_indices = filter_by_space(client, **kwargs)
    else:
        index_list = get_object_list(client, **kwargs)
        matching_indices = filter_by_timestamp(object_list=index_list, **kwargs)
    _op_loop(client, matching_indices, op=delete_index, dry_run=dry_run, **kwargs)
    logger.info(kwargs['prepend'] + 'Speficied indices deleted.')

## curator optimize [ARGS]
def optimize(client, dry_run=False, **kwargs):
    """
    Optimize indices ``older_than`` *n* ``time_unit``\s, matching the given
    ``timestring``, ``prefix``, and ``suffix`` to ``max_num_segments`` segments
    per shard.
    
    Can optionally ``delay`` a given number of seconds after each optimization.
    
    .. note::
       As this is an iterative function, default values are handled by the
       target function(s).
       
       Unless passed in `kwargs`, parameters other than ``client`` and
       ``dry_run`` will have default values assigned by the functions being
       called:
       
       :py:func:`curator.curator.get_object_list`
       
       :py:func:`curator.curator.filter_by_timestamp`
       
       :py:func:`curator.curator.optimize_index`

       These defaults are included here for documentation.

    :arg client: The Elasticsearch client connection
    :arg dry_run: If True, simulate, but do not perform the operation
    :arg older_than: Indices older than the indicated number of whole
        ``time_units`` will be operated on.
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``.
    :arg timestring: An strftime string to match the datestamp in an index name.
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg max_num_segments: Merge to this number of segments per shard.
    :arg delay: Pause *n* seconds after optimizing an index.
    :arg exclude_pattern: Exclude indices matching the provided regular
        expression.
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    """
    kwargs['prepend'] = "DRY RUN: " if dry_run else ''
    logging.info(kwargs['prepend'] + "Optimizing indices...")
    index_list = get_object_list(client, **kwargs)
    matching_indices = filter_by_timestamp(object_list=index_list, **kwargs)
    _op_loop(client, matching_indices, op=optimize_index, dry_run=dry_run, **kwargs)
    logger.info(kwargs['prepend'] + 'Optimized specified indices.')

## curator snapshot [ARGS]
def snapshot(client, dry_run=False, **kwargs):
    """
    Multiple use cases:
    
    1. Snapshot indices ``older_than`` *n* ``time_unit``\s, matching the given ``timestring``, ``prefix``, and ``suffix`` into ``repository``.
    2. Snapshot *n* ``most_recent`` indices matching ``prefix`` and ``suffix`` into ``repository``.
    3. ``delete_older_than`` *n* ``time_units`` matching the given ``timestring``, ``prefix``, and ``suffix`` from ``repository``.
    4. Snapshot ``all_indices`` (ignoring ``older_than`` and ``most_recent``).

    .. note::
       As this is an iterative function, default values are handled by the
       target function(s).
       
       Unless passed in `kwargs`, parameters other than ``client`` and
       ``dry_run`` will have default values assigned by the functions being
       called:
       
       :py:func:`curator.curator.get_object_list`
       
       :py:func:`curator.curator.filter_by_timestamp`
       
       :py:func:`curator.curator.create_snapshot`
       
       :py:func:`curator.curator.delete_snapshot`
       
       These defaults are included here for documentation.

    :arg client: The Elasticsearch client connection
    :arg dry_run: If true, simulate, but do not perform the operation
    :arg older_than: Indices older than the indicated number of whole
        ``time_unit``\s will be operated on.
    :arg most_recent: Most recent *n* indices will be operated on.
    :arg all_indices: Boolean.  Include ``_all`` indices in snapshot.
    :type all_indices: bool
    :arg delete_older_than: Snapshots older than the indicated number of whole
        ``time_unit``\s will be operated on.
    :arg time_unit: One of ``hours``, ``days``, ``weeks``, ``months``.  Default
        is ``days``.
    :arg timestring: An strftime string to match the datestamp in an index name.
    :arg snapshot_name: Override the default with this value. Defaults to `None`
    :arg snapshot_prefix: Override the default with this value. Defaults to
        ``curator-``
    :arg prefix: A string that comes before the datestamp in an index name.
        Can be empty. Wildcards acceptable.  Default is ``logstash-``.
    :arg suffix: A string that comes after the datestamp of an index name.
        Can be empty. Wildcards acceptable.  Default is empty, ``''``.
    :arg repository: The Elasticsearch snapshot repository to use
    :arg wait_for_completion: Wait (or not) for the operation
        to complete before returning.  Waits by default, i.e. Default is
        `True`
    :type wait_for_completion: bool
    :arg ignore_unavailable: Ignore unavailable shards/indices.
        Default is `False`
    :type ignore_unavailable: bool
    :arg include_global_state: Store cluster global state with snapshot.
        Default is `True`
    :type include_global_state: bool
    :arg partial: Do not fail if primary shard is unavailable. Default
        is `False`
    :type partial: bool
    :arg exclude_pattern: Exclude indices matching the provided regular
        expression.
    :arg utc_now: Used for testing.  Overrides current time with specified time.
    """
    if not 'repository' in kwargs:
        logger.error("Repository name not provided.")
        return
    kwargs['prepend'] = "DRY RUN: " if dry_run else ''
    if not 'older_than' in kwargs and not 'most_recent' in kwargs and not 'delete_older_than' in kwargs and not 'all_indices'in kwargs:
        logger.error('Expect missing argument.')
        return
    # Preserving kwargs intact for passing to _op_loop is the game here...
    all_indices       = kwargs['all_indices'] if 'all_indices' in kwargs else False
    delete_older_than = kwargs['delete_older_than'] if 'delete_older_than' in kwargs else False
    older_than        = kwargs['older_than'] if 'older_than' in kwargs else False
    most_recent       = kwargs['most_recent'] if 'most_recent' in kwargs else False
    if delete_older_than:
        logger.info(kwargs['prepend'] + "Deleting specified snapshots...")
        kwargs['older_than'] = kwargs['delete_older_than'] # Fix for delete in this case only.
        snapshot_list = client.snapshot.get(repository=kwargs['repository'], snapshot="_all")['snapshots']
        matching_snapshots = filter_by_timestamp(object_list=snapshot_list, object_type='snapshot', **kwargs)
        _op_loop(client, matching_snapshots, op=delete_snapshot, dry_run=dry_run, **kwargs)
        logger.info(kwargs['prepend'] + 'Specified snapshots deleted.')
    else:
        logger.info(kwargs['prepend'] + "Capturing snapshots of specified indices...")
        if not all_indices:
            index_list = get_object_list(client, **kwargs)
            if most_recent:
                matching_indices = index_list[-kwargs['most_recent']:]
            elif older_than:
                matching_indices = filter_by_timestamp(object_list=index_list, **kwargs)
                matching_indices = list(matching_indices)
            else:
                logger.error(kwargs['prepend'] + 'Missing argument: Must provide one of: older_than, most_recent, all_indices, delete_older_than')
                return
        else:
            matching_indices = '_all'
        if not dry_run:
            # Default `create_snapshot` behavior is to snap `_all` into a
            # snapshot named `snapshot_name` or `curator-%Y-%m-%dT%H:%M:%S`
            create_snapshot(client, indices=matching_indices, **kwargs)
        logger.info(kwargs['prepend'] + 'Snapshots captured for specified indices.')



#!/usr/bin/env python

import sys
import time
import logging
from datetime import timedelta

import elasticsearch
from curator import *

try:
    from logging import NullHandler
except ImportError:
    from logging import Handler

    class NullHandler(Handler):
        def emit(self, record):
            pass

__version__ = '2.0.0'

# Elasticsearch versions supported
version_max  = (2, 0, 0)
version_min = (1, 0, 0)

logger = logging.getLogger(__name__)

DEFAULT_ARGS = {
    'host': 'localhost',
    'url_prefix': '',
    'port': 9200,
    'ssl': False,
    'auth': None,
    'timeout': 30,
    'dry_run': False,
    'debug': False,
    'log_level': 'INFO',
    'logformat': 'default',
}

def make_parser():
    """ Creates an ArgumentParser to parse the command line options. """
    help_desc = 'Repository manager for Elasticsearch Curator.'
    try:
        import argparse
        parser = argparse.ArgumentParser(description=help_desc)
        parser.add_argument('-v', '--version', action='version', version='%(prog)s '+__version__)
    except ImportError:
        print('{0} requires module argparse.  Try: pip install argparse'.format(sys.argv[0]))
        sys.exit(1)
    parser.add_argument('--host', help='Elasticsearch host. Default: localhost', default=DEFAULT_ARGS['host'])
    parser.add_argument('--url_prefix', help='Elasticsearch http url prefix. Default: none', default=DEFAULT_ARGS['url_prefix'])
    parser.add_argument('--port', help='Elasticsearch port. Default: 9200', default=DEFAULT_ARGS['port'], type=int)
    parser.add_argument('--ssl', help='Connect to Elasticsearch through SSL. Default: false', action='store_true', default=DEFAULT_ARGS['ssl'])
    parser.add_argument('--auth', help='Use Basic Authentication ex: user:pass Default: None', default=DEFAULT_ARGS['auth'])
    parser.add_argument('-t', '--timeout', help='Connection timeout in seconds. Default: 30', default=DEFAULT_ARGS['timeout'], type=int)
    parser.add_argument('-n', '--dry-run', action='store_true', help='If true, does not perform any changes to the Elasticsearch indices.',
                    default=DEFAULT_ARGS['dry_run'])
    parser.add_argument('-D', '--debug', dest='debug', action='store_true', help='Debug mode', default=DEFAULT_ARGS['debug'])
    parser.add_argument('--loglevel', dest='log_level', action='store', help='Log level', default=DEFAULT_ARGS['log_level'], type=str)
    parser.add_argument('-l', '--logfile', dest='log_file', help='log file', type=str)
    parser.add_argument('--logformat', dest='logformat', help='Log output format [default|logstash]. Default: default', default=DEFAULT_ARGS['logformat'], type=str)

    # Command sub_parsers
    subparsers = parser.add_subparsers(title='Commands', dest='command', description='Select one of the following commands:',
                    help='Run: ' + sys.argv[0] + ' COMMAND --help for command-specific help.')

    # 'fs' Repository creation
    parser_fs = subparsers.add_parser('create_fs', help='Create an "fs" type repository')
    parser_fs.set_defaults(func=create_repository, repo_type='fs')
    parser_fs.add_argument('-r', '--repository', dest='repository', required=True, help='Repository name', type=str)
    parser_fs.add_argument('--location', dest='location', action='store', type=str, required=True,
                    help='Shared file-system location. Must match remote path, & be accessible to all master & data nodes')
    parser_fs.add_argument('--disable-compression', dest='compress', action='store_false', default=True,
                    help='Disable compression (enabled by default)')
    parser_fs.add_argument('--concurrent_streams', dest='concurrent_streams', action='store', type=int, default=None,
                        help='Number of streams (per node) preforming snapshot. Default: 5')
    parser_fs.add_argument('--chunk_size', dest='chunk_size', action='store', type=str, default=None,
                        help='Chunk size, e.g. 1g, 10m, 5k. Default is unbounded.')
    parser_fs.add_argument('--max_restore_bytes_per_sec', dest='max_restore_bytes_per_sec', action='store', type=str, default=None,
                        help='Throttles per node restore rate (per second). Default: 20mb')
    parser_fs.add_argument('--max_snapshot_bytes_per_sec', dest='max_snapshot_bytes_per_sec', action='store', type=str, default=None,
                        help='Throttles per node snapshot rate (per second). Default: 20mb')

    # 's3' Repository creation
    parser_s3 = subparsers.add_parser('create_s3', help='Create an \'s3\' type repository')
    parser_s3.set_defaults(func=create_repository, repo_type='s3')
    parser_s3.add_argument('-r', '--repository', dest='repository', required=True, help='Repository name', type=str)
    parser_s3.add_argument('--bucket', dest='bucket', action='store', type=str, required=True,
                        help='S3 bucket name')
    parser_s3.add_argument('--region', dest='region', action='store', type=str, default=None,
                        help='S3 region. Defaults to US Standard')
    parser_s3.add_argument('--base_path', dest='base_path', action='store', type=str, default=None,
                        help='S3 base path. Defaults to root directory.')
    parser_s3.add_argument('--access_key', dest='access_key', action='store', type=str, default=None,
                        help='S3 access key. Defaults to value of cloud.aws.access_key')
    parser_s3.add_argument('--secret_key', dest='secret_key', action='store', type=str, default=None,
                        help='S3 secret key. Defaults to value of cloud.aws.secret_key')
    parser_s3.add_argument('--disable-compression', dest='compress', action='store_false', default=True,
                        help='Disable compression (enabled by default)')
    parser_s3.add_argument('--concurrent_streams', dest='concurrent_streams', action='store', type=int, default=None,
                        help='Number of streams (per node) preforming snapshot. Default: 5')
    parser_s3.add_argument('--chunk_size', dest='chunk_size', action='store', type=str, default=None,
                        help='Chunk size, e.g. 1g, 10m, 5k. Default is unbounded.')
    parser_s3.add_argument('--max_restore_bytes_per_sec', dest='max_restore_bytes_per_sec', action='store', type=str, default=None,
                    help='Throttles per node restore rate (per second). Default: 20mb')
    parser_s3.add_argument('--max_snapshot_bytes_per_sec', dest='max_snapshot_bytes_per_sec', action='store', type=str, default=None,
                        help='Throttles per node snapshot rate (per second). Default: 20mb')

    # Delete repository
    parser_delete = subparsers.add_parser('delete', help='Delete named repository')
    parser_delete.set_defaults(func=delete_repository)
    parser_delete.add_argument('-r', '--repository', dest='repository', required=True, help='Repository name', type=str)


        # Show registered repositories
    parser_show = subparsers.add_parser('show', help='Show all registered repositories')
    parser_show.set_defaults(func=show)

    return parser

class Whitelist(logging.Filter):
    def __init__(self, *whitelist):
        self.whitelist = [logging.Filter(name) for name in whitelist]

    def filter(self, record):
        return any(f.filter(record) for f in self.whitelist)

def show(client, **kwargs):
    for repository in sorted(get_repository(client, '_all').keys()):
        print('{0}'.format(repository))
    sys.exit(0)

def check_version(client):
    """
    Verify version is within acceptable range.  Exit with error if it is not.
    
    :arg client: The Elasticsearch client connection
    """
    version_number = get_version(client)
    logger.debug('Detected Elasticsearch version {0}'.format(".".join(map(str,version_number))))
    if version_number >= version_max or version_number < version_min:
        print('Expected Elasticsearch version range > {0} < {1}'.format(".".join(map(str,version_min)),".".join(map(str,version_max))))
        print('ERROR: Incompatible with version {0} of Elasticsearch.  Exiting.'.format(".".join(map(str,version_number))))
        sys.exit(1)

def create_repository(client, dry_run=False, repository='', **kwargs):
    """Create repository with repository and body settings"""
    if not repository:
        logger.error("No repository specified.")
        sys.exit(1)
    if not dry_run:
        try:
            body = create_repo_body(**kwargs)
            logging.info("Checking if repository {0} already exists...".format(repository))
            result = get_repository(client, repository=repository)
            if not result:
                logging.info("Repository {0} not found. Continuing...".format(repository))
                client.snapshot.create_repository(repository=repository, body=body)
            elif result is not None and repository not in result and not kwargs['dry_run']:
                logging.info("Repository {0} not found. Continuing...".format(repository))
                client.snapshot.create_repository(repository=repository, body=body)
            else:
                logger.error("Unable to create repository {0}.  A repository with that name already exists.".format(repository))
                sys.exit(0)
        except:
            logger.error("Unable to create repository {0}.  Check logs for more information.".format(repository))
            return False
        logger.info("Repository {0} creation initiated...".format(repository))
        test_result = get_repository(client, repository)
        if repository in test_result:
            logger.info("Repository {0} creation validated.".format(repository))
            return True
        else:
            logger.error("Repository {0} failed validation...".format(repository))
            return False
    else:
        logger.info("Would have attempted to create repository {0}".format(repository))

def delete_repository(client, repository='', dry_run=False, **kwargs):
    """
    Delete indicated repository.
    """
    if not dry_run:
        try:
            logger.info('Deleting repository {0}...'.format(repository))
            return client.snapshot.delete_repository(repository=repository)
        except elasticsearch.NotFoundError as e:
            logger.error("Error: {0}".format(e))
            return False
    else:
        logger.info("Would have attempted to delete repository {0}".format(repository))

def create_repo_body(repo_type='fs',
                     compress=True, concurrent_streams=None, chunk_size=None, max_restore_bytes_per_sec=None, max_snapshot_bytes_per_sec=None,
                     location=None,
                     bucket=None, region=None, base_path=None, access_key=None, secret_key=None, **kwargs):
    """Create the request body for creating a repository"""
    argdict = locals()
    body = {}
    body['type'] = argdict['repo_type']
    body['settings'] = {}
    settings = []
    maybes   = ['compress', 'concurrent_streams', 'chunk_size', 'max_restore_bytes_per_sec', 'max_snapshot_bytes_per_sec']
    s3      = ['bucket', 'region', 'base_path', 'access_key', 'secret_key']

    settings += [i for i in maybes if argdict[i]]
    # Type 'fs'
    if argdict['repo_type'] == 'fs':
        settings.append('location')
    # Type 's3'
    if argdict['repo_type'] == 's3':
        settings += [i for i in s3 if argdict[i]]
    for k in settings:
        body['settings'][k] = argdict[k]
    return body

def main():
    start = time.time()

    parser = make_parser()
    arguments = parser.parse_args()

    # Do not log and force dry-run if we opt to show indices.
    if arguments.command == 'show':
        import os
        arguments.log_file = os.devnull
        arguments.dry_run = True

    # Setup logging
    if arguments.debug:
        numeric_log_level = logging.DEBUG
        format_string = '%(asctime)s %(levelname)-9s %(name)22s %(funcName)22s:%(lineno)-4d %(message)s'
    else:
        numeric_log_level = getattr(logging, arguments.log_level.upper(), None)
        format_string = '%(asctime)s %(levelname)-9s %(message)s'
        if not isinstance(numeric_log_level, int):
            raise ValueError('Invalid log level: %s' % arguments.log_level)
    
    date_string = None
    if arguments.logformat == 'logstash':
        os.environ['TZ'] = 'UTC'
        time.tzset()
        format_string = '{"@timestamp":"%(asctime)s.%(msecs)03dZ", "loglevel":"%(levelname)s", "name":"%(name)s", "function":"%(funcName)s", "linenum":"%(lineno)d", "message":"%(message)s"}'
        date_string = '%Y-%m-%dT%H:%M:%S'
        
    logging.basicConfig(level=numeric_log_level,
                        format=format_string,
                        datefmt=date_string,
                        stream=open(arguments.log_file, 'a') if arguments.log_file else sys.stderr)
    
    
    # Filter out logging from Elasticsearch and associated modules by default
    if not arguments.debug:
        for handler in logging.root.handlers:
            handler.addFilter(Whitelist('root', '__main__', 'curator.es_repo_mgr'))
    logging.info("Job starting...")

    # Setting up NullHandler to handle nested elasticsearch.trace Logger instance in elasticsearch python client
    logging.getLogger('elasticsearch.trace').addHandler(NullHandler())

    if arguments.dry_run:
        logger.info('DRY RUN MODE.  No changes will be made.')

    client = elasticsearch.Elasticsearch(host=arguments.host, http_auth=arguments.auth, port=arguments.port, url_prefix=arguments.url_prefix, timeout=arguments.timeout, use_ssl=arguments.ssl)

    check_version(client)

    # Execute the command specified in the arguments
    argdict = arguments.__dict__
    logging.debug("argdict = {0}".format(argdict))
    arguments.func(client, **argdict)
    
    logger.info('Done in {0}.'.format(timedelta(seconds=time.time()-start)))
    
if __name__ == '__main__':
    main()
